/*--------------------------------------------------------------------------
 *  Copyright 2007 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// utgb-core Project
//
// EGTXMLReader.java
// Since: Dec 21, 2007
//
// $URL: http://svn.utgenome.org/utgb/tags/utgb/utgb-core/utgb-core-1.0.3/src/main/java/org/utgenome/format/egt/EGTXMLReader.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.format.egt;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.Reader;

import org.utgenome.UTGBException;
import org.xerial.core.XerialErrorCode;
import org.xerial.db.DBException;
import org.xerial.db.sql.SQLExpression;
import org.xerial.db.sql.SQLUtil;
import org.xerial.db.sql.sqlite.SQLiteAccess;
import org.xerial.util.bean.BeanException;
import org.xerial.util.bean.BeanUtil;
import org.xerial.util.cui.OptionHandler;
import org.xerial.util.cui.OptionParser;
import org.xerial.util.cui.OptionParserException;
import org.xerial.util.log.Logger;

public class EGTXMLReader {

	static abstract class GeneDBGenerator {
		final SQLiteAccess sqlite;

		public GeneDBGenerator(String dbFileName) throws DBException {
			File f;
			if ((f = new File(dbFileName)).exists()) {
				f.delete();
			}
			sqlite = new SQLiteAccess(dbFileName);
		}

		public abstract void insertGene(Gene gene) throws DBException;

		public abstract void createIndexes() throws DBException;

		public void beginTransaction() throws DBException {
			sqlite.update("begin transaction", false);
		}

		public void commit() throws DBException {
			sqlite.update("commit", false);
		}
	}

	static class CompactDBGenerator extends GeneDBGenerator {
		int geneCount = 0;

		public CompactDBGenerator(String dbFileName) throws DBException {
			super(dbFileName);

			sqlite.update("create table gene ( " + "id integer primary key not null, " + "target string, " + "start integer, " + "end integer, "
					+ "strand string, " + "name string)", false);

			sqlite.update("create table gene_info ( " + "id integer primary key not null, " + "url string)", false);

			sqlite.update("create table exon ( " + "gene_id integer, " + "id integer, " + "start integer, " + "end integer, "
					+ "primary key (gene_id, id, start, end))", false);

		}

		public void insertGene(Gene gene) throws DBException {
			int geneID = geneCount++;
			sqlite.update(SQLExpression.fillTemplate("insert into gene values($1, $2, $3, $4, $5, $6)", geneID, SQLUtil.singleQuote(gene.getTarget()), gene
					.getStart(), gene.getEnd(), SQLUtil.singleQuote(gene.getStrand()), SQLUtil.singleQuote(gene.getName())), false);

			sqlite.update(SQLExpression.fillTemplate("insert into gene_info values($1, $2)", geneID, SQLUtil.singleQuote(gene.getUrl())), false);

			int exonID = 1;
			for (Exon e : gene.getExon()) {
				sqlite.update(SQLExpression.fillTemplate("insert into exon values($1, $2, $3, $4)", geneID, exonID++, e.getStart(), e.getEnd()), false);
			}

		}

		@Override
		public void createIndexes() throws DBException {

		}

	}

	static class MinimalDBGenerator extends GeneDBGenerator {
		int geneCount = 0;

		public MinimalDBGenerator(String dbFileName) throws DBException {
			super(dbFileName);

			sqlite.update("create table gene ( " + "id integer primary key not null, " + "target string, " + "start integer, " + "end integer, "
					+ "strand string, " + "name string)", false);

			/*
			sqlite.update(
					"create table gene_info ( " +
					"id integer primary key not null, " +
					"url string)");
			
			
			sqlite.update(
					"create table exon ( " +
					"gene_id integer, " +
					"id integer, " +
					"start integer, " +
					"end integer, " +
					"primary key (gene_id, id, start, end))");
					*/

		}

		public void insertGene(Gene gene) throws DBException {
			int geneID = geneCount++;
			sqlite.update(SQLExpression.fillTemplate("insert into gene values($1, $2, $3, $4, $5, $6)", geneID, SQLUtil.singleQuote(gene.getTarget()), gene
					.getStart(), gene.getEnd(), SQLUtil.singleQuote(gene.getStrand()), SQLUtil.singleQuote(gene.getName())), false);

			/*
			sqlite.update(SQLExpression.fillTemplate(
					"insert into gene_info values($1, $2)", 
					geneID,
					SQLUtil.singleQuote(gene.getUrl())));
			
			int exonID = 1;
			for(Exon e : gene.getExon())
			{
				sqlite.update(SQLExpression.fillTemplate(
						"insert into exon values($1, $2, $3, $4)",
						geneID,
						exonID++,
						e.getStart(),
						e.getEnd()));
			}
			*/

		}

		@Override
		public void createIndexes() throws DBException {

		}

	}

	static class WithIndexDBGenerator extends MinimalDBGenerator {
		public WithIndexDBGenerator(String dbName) throws DBException {
			super(dbName);

			sqlite.update("create index gene_index on gene (target, start)", false);

		}
	}

	private static enum TableType {
		COMPACT, MINIMAL, WITHINDEX, DENORMALIZED, VERBOSE
	}

	public static enum Opt {
		HELP, DBTYPE, READ_STDIN, DBFILENAME, USEMEMORYDB
	}

	private static Logger _logger = Logger.getLogger(EGTXMLReader.class);
	private TableType tableType = TableType.COMPACT;
	private GeneDBGenerator dbGenerator;
	private String dbName = "egt.db";
	private boolean useMemoryDB = false;

	public static void main(String[] args) {
		try {
			EGTXMLReader reader = new EGTXMLReader(args);
		}
		catch (OptionParserException e) {
			System.err.println(e);
		}
	}

	public EGTXMLReader(String[] args) throws OptionParserException {
		OptionParser<Opt> parser = new OptionParser<Opt>();
		parser.addOption(Opt.HELP, "h", "help", "display help message");
		parser.addOptionWithArgument(Opt.DBTYPE, "t", "type", "DBTYPE", "0:compact, 1:minimal, 2:denomalized, 3:verbose", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				int v = parser.getIntValue(Opt.DBTYPE);
				tableType = TableType.values()[v];
			}
		});
		parser.addOptionWithArgument(Opt.DBFILENAME, "o", "out", "DBFILE", "output db file name", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				dbName = parser.getValue(Opt.DBFILENAME);
			}
		});

		parser.addOption(Opt.READ_STDIN, "c", "stream", "read from standard input");
		parser.addOption(Opt.USEMEMORYDB, "m", "memory", "use memory db", new OptionHandler<Opt>() {

			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				useMemoryDB = true;
			}

		});

		parser.parse(args);

		if (parser.isSet(Opt.HELP)) {
			System.out.println(parser.helpMessage());
			return;
		}

		try {

			switch (tableType) {
			case MINIMAL:
				dbGenerator = new MinimalDBGenerator(dbName);
				break;
			case WITHINDEX:
				dbGenerator = new WithIndexDBGenerator(dbName);
				break;
			case COMPACT:
			default:
				dbGenerator = new CompactDBGenerator(dbName);
				break;
			}

			if (parser.isSet(Opt.READ_STDIN)) {
				load(new InputStreamReader(System.in));
			}
			else {
				if (parser.getArgumentLength() < 1)
					throw new OptionParserException(XerialErrorCode.INVALID_INPUT, "no xml file is given");

				Reader xmlReader = null;
				try {
					xmlReader = new BufferedReader(new FileReader(parser.getArgument(0)));
					load(xmlReader);
				}
				catch (FileNotFoundException e) {
					throw new UTGBException(e);
				}
				finally {
					if (xmlReader != null)
						xmlReader.close();
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

	}

	public void load(Reader xmlReader) throws UTGBException, BeanException, DBException {
		if (dbGenerator == null)
			throw new UTGBException("db generator is null");

		beginTime = System.currentTimeMillis();

		BeanUtil.populateBeanWithXML(this, xmlReader);
		dbGenerator.commit();
	}

	private int geneCount = 0;
	private long beginTime = 0;

	public void addGene(Gene gene) {
		try {
			dbGenerator.insertGene(gene);
			if ((geneCount % 10000) == 0) {
				_logger.info("inserted: " + geneCount + "\t " + (System.currentTimeMillis() - beginTime) / 1000.0 + " sec.");
			}
			geneCount++;
		}
		catch (DBException e) {
			_logger.error(e);
		}
	}

}
