/*--------------------------------------------------------------------------
 *  Copyright 2007 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// GenomeBrowser Project
//
// UTGB.java
// Since: May 29, 2007
//
// $URL: http://svn.utgenome.org/utgb/tags/utgb/utgb-core/utgb-core-1.0.3/src/main/java/org/utgenome/gwt/utgb/server/UTGBMaster.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.gwt.utgb.server;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;

import org.utgenome.UTGBErrorCode;
import org.utgenome.UTGBException;
import org.utgenome.gwt.utgb.server.config.ConnectionInfo;
import org.utgenome.gwt.utgb.server.config.DatabaseInfo;
import org.utgenome.gwt.utgb.server.config.ImportStmt;
import org.utgenome.gwt.utgb.server.config.TrackConfiguration;
import org.xerial.core.XerialException;
import org.xerial.db.DBException;
import org.xerial.db.sql.DatabaseAccess;
import org.xerial.db.sql.sqlite.SQLiteAccess;
import org.xerial.util.bean.BeanUtil;
import org.xerial.util.log.Logger;

/**
 * UTGB contains global variables that should be shared throughout the entire server-side programs
 */
public class UTGBMaster implements ServletContextListener {
	private static Logger _logger = Logger.getLogger(UTGBMaster.class);
	private static UTGBMaster _instance = new UTGBMaster();

	// private String contextPath = "/";
	private ClassLoader trackClassLoader = null;
	private TrackConfiguration config = null;
	private HashMap<String, Object> _glovalVariableHolder = new HashMap<String, Object>();
	private Properties _property = new Properties();
	private HashMap<String, DatabaseAccess> _dbAccessTable = new HashMap<String, DatabaseAccess>();

	public static UTGBMaster getInstance() {
		return (UTGBMaster) _instance;
	}

	public static String getTrackConfigProperty(String key, String defaultValue) {
		return _instance.getTrackConfigProeprtyInternal(key, defaultValue);
	}

	private String getTrackConfigProeprtyInternal(String key, String defaultValue) {
		try {
			if (config == null)
				config = loadTrackConfig();

			if (config == null)
				return defaultValue;

			return config.getProperty(key, defaultValue);
		}
		catch (Exception e) {
			_logger.error(e);
			return defaultValue;
		}

	}

	public static String getProperty(String key, String defaultValue) {
		return _instance._property.getProperty(key, defaultValue);
	}

	public static String getProperty(String key) {
		return getProperty(key, "");
	}

	public static Object getVariable(String name) {
		return _instance._glovalVariableHolder.get(name);
	}

	public static void setVariable(String name, Object value) {
		_instance._glovalVariableHolder.put(name, value);
	}

	public void loadProject(ClassLoader trackProjetClassLoader) {
		try {
			// load track config
			config = loadTrackConfig();

			// load the default action package
			DefaultRequestMap.loadActionPackage(trackProjetClassLoader, config.getPackage() + ".app", "");

			// load the imported packages
			for (ImportStmt importStatement : config.getImportList()) {
				String actionPackage = importStatement.getActionPackage();
				String actionPrefix = importStatement.getAlias();

				_logger.info("import " + actionPackage + " (alias = " + actionPrefix + ")");
				DefaultRequestMap.loadActionPackage(trackProjetClassLoader, actionPackage, actionPrefix);
			}

			// search database resources specified in the track-config.xml

			// search SQLite database files
			for (DatabaseInfo dbInfo : config.getDatabaseList()) {
				_logger.info("database id: " + dbInfo.getId());
				for (ConnectionInfo connectionInfo : dbInfo.getConnectionList()) {
					_logger.info("connection: " + connectionInfo.getAddress());
				}
			}

		}
		catch (UTGBException e) {
			_logger.error(e);
		}
	}

	public static TrackConfiguration loadTrackConfig() throws UTGBException {

		String projectRootFolder = (String) getVariable("projectRoot");
		if (projectRootFolder == null)
			throw new UTGBException("not in the project root folder, or a file 'config/track-config.xml' is not found");

		try {
			BufferedReader reader = new BufferedReader(new FileReader(new File(projectRootFolder, "config/track-config.xml")));
			TrackConfiguration config = BeanUtil.createXMLBean(TrackConfiguration.class, reader);
			return config;
		}
		catch (XerialException e) {
			throw new UTGBException("syntax error in track-config.xml file: " + e.getMessage());
		}
		catch (IOException e) {
			throw new UTGBException("failed to load config/track-config.xml " + e.getMessage());
		}
	}

	/**
	 * recursively scan the folder for the target files
	 * 
	 * @param dir
	 * @param filter
	 * @param fileList
	 */
	private void scanFile(File dir, FileFilter filter, List<File> fileList) {
		if (dir.exists() && dir.isDirectory()) {
			for (File f : dir.listFiles()) {
				if (f.isDirectory())
					scanFile(f, filter, fileList);
				else {
					if (filter.accept(f))
						fileList.add(f);
				}
			}
		}

	}

	public static String getContextProperty(String key, String defaultValue) {
		String value = null;
		try {
			// read the environment variable from the JNDI,
			// the parmeters of which are configured in META-INF/context.xml in
			// the Tomcat server.
			Context initContext = new InitialContext();
			Context envContext = (Context) initContext.lookup("java:comp/env");
			value = (String) envContext.lookup(key);
			System.setProperty(key, value);
			return value;
		}
		catch (NamingException e) {
			_logger.warn("error while reading an environment context: " + key + ". " + e.getMessage());
			_logger.warn("using the default value: " + defaultValue);
			System.setProperty(key, defaultValue);
			return defaultValue;
		}
	}

	public static DatabaseAccess getDatabaseAccess(String databaseID) throws UTGBException {
		return getInstance().getDatabaseAccessInternal(databaseID);
	}

	protected DatabaseAccess getDatabaseAccessInternal(String databaseID) throws UTGBException {
		if (_dbAccessTable.containsKey(databaseID))
			return _dbAccessTable.get(databaseID);

		String projectRootFolder = (String) getVariable("projectRoot");
		if (projectRootFolder == null)
			throw new UTGBException("not in the project root folder, or a file 'config/track-config.xml' is not found");

		if (config == null) {
			config = loadTrackConfig();

		}

		DatabaseInfo dbInfo = config.getDatabase(databaseID);
		if (dbInfo == null)
			throw new UTGBException(UTGBErrorCode.DatabaseError, "no database ID " + databaseID + " was found");

		if (dbInfo.getConnectionList().size() <= 0)
			throw new UTGBException(UTGBErrorCode.DatabaseError, "no connection was found for the database ID = " + databaseID);

		// TODO select one of the appropriate connections
		ConnectionInfo conn = dbInfo.getConnectionList().get(0);
		DatabaseAccess dbAccess = JDBCService.getDatabaseAccess(projectRootFolder, conn);
		_dbAccessTable.put(databaseID, dbAccess);
		return dbAccess;
	}

	public void initialize(ServletContextEvent event, ClassLoader classLoaderOfTheProject) {
		this.trackClassLoader = classLoaderOfTheProject;

		// SQLiteJDBCLoader.initialize(true);

		setVariable("query", new HashMap<String, SQLiteAccess>());

		// scan resource folder
		String configFolderName = getContextProperty("projectRoot", new File(".").getAbsolutePath());
		setVariable("projectRoot", configFolderName);
		_logger.info("project root folder: " + configFolderName);

		loadProject(trackClassLoader);
	}

	public void dispose(ServletContextEvent event) {

		for (String key : _dbAccessTable.keySet()) {
			DatabaseAccess query = _dbAccessTable.get(key);
			if (query != null)
				try {
					// close JDBC connections
					query.dispose();
				}
				catch (DBException e) {
					_logger.warn(e);
				}
		}
	}

	public void contextDestroyed(ServletContextEvent event) {
		dispose(event);
	}

	public void contextInitialized(ServletContextEvent event) {
		initialize(event, Thread.currentThread().getContextClassLoader());
	}

	public static BufferedReader openServletReader(HttpServletRequest request, String path) throws ServletException, IOException {
		if (!path.startsWith("/"))
			path = "/" + path;

		String servletURL = request.getLocalAddr() + ":" + request.getLocalPort() + request.getContextPath() + path;
		URL url;
		try {
			url = new URL("http://" + servletURL);
			return new BufferedReader(new InputStreamReader(url.openStream()));
		}
		catch (MalformedURLException e) {
			throw new ServletException(e);
		}
	}

}
